from __future__ import annotations
from dataclasses import dataclass, field
from abc import ABC, abstractmethod
from itertools import product, count, repeat
import string
import random
from .node import Node
from .util import *

# TODO: verbosity in parsing methods?

@dataclass
class Element(ABC):
    """Represents a grammatical element."""

    # TODO: figure out how to not have to duplicate all the default value fields in the element subclasses

    def _generate(self, grammar: Grammar, depth: int = 0, verbose: bool = False):
        """Generate all possible strings this element can match."""

        if verbose:
            print(f'Generation depth = {depth}.')

        if depth < grammar.max_recursions:
            if verbose and self.generation_override:
                print(f'Generation override = {self.generation_override!r}.')
            return [self.generation_override] if self.generation_override else self.generate(grammar, depth, verbose)

        if verbose:
            print(f'Generation depth too high; stopping generation here.')

        return iter(())

    @abstractmethod
    def generate(self, grammar: Grammar, verbose: bool = False):
        """Subclass generation method."""

        ...

    @abstractmethod
    def parse(self, grammar: Grammar, rule: str, string: str, position: int = 0, verbose: bool = False):
        """Generate all possible parse trees."""

        ...

@dataclass
class Terminal(Element):
    """Represents a terminal string element."""

    # literal string to use as the terminal element
    string: str
    generation_override: str = None
    label: str = None

    def generate(self, grammar: Grammar, depth: int = 0, verbose: bool = False):
        """Generate the terminal."""

        if verbose:
            print(f'Generating terminal {self.string!r}.')

        yield self.string

    def parse(self, grammar: Grammar, rule: str, string: str, position: int = 0, verbose: bool = False):
        """Parse the terminal."""

        region = slice(position, position + len(self.string))
        if string[region] == self.string:
            yield Node(grammar, rule, string, region, label=self.label)

    def __str__(self):
        label = f'{self.label}:' if self.label else ''
        return f'{label}"{self.string}"'

@dataclass
class Substitution(Element):
    """Represents a stand-in for another grammar rule."""

    # the name of the rule in the grammar
    name: str
    generation_override: str = None
    label: str = None

    def generate(self, grammar: Grammar, depth: int = 0, verbose: bool = False):
        """Generate the non-terminal."""

        if verbose:
            print(f'Generating substitution rule {self.name!r}.')

        return grammar.rules[self.name]._generate(grammar, depth, verbose)

    def parse(self, grammar: Grammar, rule: str, string: str, position: int = 0, verbose: bool = False):
        """Parse the non-terminal."""

        return label_node(grammar.rules[self.name].parse(grammar, self.name, string, position, verbose), self.label)

    def __str__(self):
        label = f'{self.label}:' if self.label else ''
        return f'{label}{self.name}'

@dataclass
class Sequence(Element):
    """Represents a sequence of elements."""

    # the sequence
    elements: list
    generation_override: str = None
    label: str = None

    def generate(self, grammar: Grammar, depth: int = 0, verbose: bool = False):
        """Generate all possible sequences."""

        if verbose:
            print(f'Generating sequence.')

        return generate_product(grammar, lambda: (element._generate(grammar, depth + 1, verbose) for element in self.elements), verbose)

    def parse(self, grammar: Grammar, rule: str, string: str, position: int = 0, verbose: bool = False):
        """Parse the sequence."""

        return label_node(parse_sequence(self.elements, grammar, rule, string, position, verbose), self.label)

    def __str__(self):
        return ' '.join(map(str, self.elements))

@dataclass
class Choice(Element):
    """Represents a choice among elements."""

    # the possible choices
    elements: list
    generation_override: str = None
    label: str = None

    def generate(self, grammar: Grammar, depth: int = 0, verbose: bool = False):
        """Generate all possible choices."""

        if verbose:
            print(f'Generating choice.')

        for element in self.elements:
            for string in element._generate(grammar, depth + 1, verbose):
                yield string

    def parse(self, grammar: Grammar, rule: str, string: str, position: int = 0, verbose: bool = False):
        """Parse the choice."""

        return label_node(parse_choice(self.elements, grammar, rule, string, position, verbose), self.label)

    def __str__(self):
        return ' | '.join(map(str, self.elements))

@dataclass
class Option(Element):
    """Represents an optional element."""

    # the optional element
    element: Element
    generation_override: str = None
    label: str = None

    def generate(self, grammar: Grammar, depth: int = 0, verbose: bool = False):
        """Generate the optional element."""

        if verbose:
            print(f'Generating option.')

        yield ""
        for string in self.element._generate(grammar, depth + 1, verbose):
            yield string

    def parse(self, grammar: Grammar, rule: str, string: str, position: int = 0, verbose: bool = False):
        """Parse the optional element."""

        yield Node(grammar, rule, string, slice(position, position), label=self.label)
        for node in label_node(self.element.parse(grammar, rule, string, position, verbose), self.label):
            yield node
                
    def __str__(self):
        return f'[ {self.element} ]'

@dataclass
class Repetition(Element):
    """Represents an n-times repeated element."""

    # the element to be repeated
    element: Element
    generation_override: str = None
    label: str = None

    def generate(self, grammar: Grammar, depth: int = 0, verbose: bool = False):
        """Generate the repeated element."""

        if verbose:
            print(f'Generating repetition.')

        yield ""
        for n in range(1, grammar.max_repetitions):
            for prod in generate_product(grammar, lambda: (self.element._generate(grammar, depth + 1, verbose) for _ in range(n)), verbose):
                yield prod

    def parse(self, grammar: Grammar, rule: str, string: str, position: int = 0, verbose: bool = False):
        """Parse the repeated element."""

        yield Node(grammar, rule, string, slice(position, position), label=self.label)
        for n in count(1):
            elements = list(repeat(self.element, n))
            nodes = parse_sequence(elements, grammar, rule, string, position, verbose)
            m = 0
            for node in nodes:
                m += 1
                if self.label:
                    yield Node(grammar, rule, string, node.region, [node], label=self.label)
                else:
                    yield node
            if m == 0:
                break

    def __str__(self):
        return f'{{ {self.element} }}'

@dataclass
class StringLiteralCharacter(Element):
    """Specialized element that accepts literal string characters.

    This includes all unicode characters, excludes the unescaped string delimiter ("), escape character (\), and includes all escaped characters.
    """

    generation_override: str = None
    label: str = None

    delimiter: str = '"'
    escape_character: str = '\\'

    def escape(self, string: str) -> str:
        """Escape every character in a given string."""

        return ''.join([self.escape_character + c for c in string])

    def generate(self, grammar: Grammar, depth: int = 0, verbose: bool = False):
        """Generate the character."""

        characters = list(string.printable)
        characters[characters.index(self.escape_character)] = self.escape(self.escape_character)
        characters[characters.index(self.delimiter)] = self.escape(self.delimiter)
        # TODO: maybe add some common escape sequences?
        return random.choice(characters)

    def parse(self, grammar: Grammar, rule: str, string: str, position: int = 0, verbose: bool = False):
        """Parse the character."""

        character = string[position]
        region = None
        if character == self.escape_character:
            region = slice(position + 1, position + 2)
        elif character != self.delimiter:
            region = slice(position, position + 1)
        if region:
            yield Node(grammar, rule, string, region, label=self.label)

    # NOTE: there is no __str__ because this is a special element that is only used internally (for now lol)



"""Expression parser and generator using defined grammars."""

from __future__ import annotations
from dataclasses import dataclass, field
import string
from .node import Node
from .element import *

@dataclass
class Grammar:
    """A set of production rules describing a context-free grammar."""

    # the set of production rules
    rules: dict

    # maximum product length, repetitions, and recursions for generation
    max_products:    int = 8
    max_repetitions: int = 4
    max_recursions:  int = 8

    def generate(self, rule_name: str = 'main', verbose: bool = False):
        """Generate all possible expressions matching a rule in the grammar."""

        if verbose:
            print(f'Generating with starting rule {rule_name!r}:')

        return self.rules[rule_name]._generate(self, verbose=verbose)

    def parse(self, string: str, rule_name: str = 'main', verbose: bool = False):
        """Generate all possible parse trees matching a string according to a rule in the grammar."""

        # TODO: figure out a system for reporting likely syntax errors?

        if verbose:
            print(f'Parsing with starting rule {rule_name!r}:')

        return self.rules[rule_name].parse(self, rule_name, string, verbose=verbose)

    def parse_complete(self, string: str, rule_name: str = 'main', verbose: bool = False):
        """Parse a string but filter out any incomplete parse trees."""

        return filter(lambda tree: tree.is_complete, self.parse(string, rule_name, verbose))

    @classmethod
    def parse_bnf(cls, source: str):
        """Return a grammar described by a given BNF source string."""

        # TODO: maybe extract this grammar out so that it doesnt have to be rebuild everytime?
        # TODO: support for comments and import statements
        # TODO: support for convenience elements (whitespace, string literals, numbers, etc)

        # create the BNF grammar itself
        bnf = Grammar({
            'string_contents':     Repetition(StringLiteralCharacter()),
            'optional_whitespace': Repetition(Choice([Terminal(char) for char in string.whitespace]), generation_override=' '),

            # identifier names can contain alphanumeric characters and underscores but must not start with a number
            'identifier': Sequence([Choice([Terminal(char) for char in string.ascii_letters + '_']), Repetition(Choice([Terminal(char) for char in string.ascii_letters + string.digits + '_']))]),

            # the production rules
            'rule': Sequence([Substitution('identifier', label='name'), Substitution('optional_whitespace'),
                              Option(Sequence([Terminal('"'), Substitution('string_contents', label='override'), Terminal('"'), Substitution('optional_whitespace')])),
                              Terminal('='), Substitution('optional_whitespace'),
                              Substitution('expression', label='expression'), Substitution('optional_whitespace'),
                              Terminal('.')]),

            # right hand expression
            'string':     Sequence([Terminal('"'), Substitution('string_contents', label='contents'), Terminal('"')]),
            'repetition': Sequence([Terminal('{'), Substitution('optional_whitespace'), Substitution('expression', label='expression'), Substitution('optional_whitespace'), Terminal('}')]),
            'option':     Sequence([Terminal('['), Substitution('optional_whitespace'), Substitution('expression', label='expression'), Substitution('optional_whitespace'), Terminal(']')]),
            'item':       Sequence([Option(Sequence([Substitution('identifier', label='label'), Substitution('optional_whitespace'), Terminal(':'), Substitution('optional_whitespace')])), Choice([Substitution('string'), Substitution('identifier'), Substitution('option'), Substitution('repetition')], label='contents')]),
            'sequence':   Sequence([Substitution('item', label='item'), Repetition(Sequence([Choice([Terminal(char) for char in string.whitespace], generation_override=' '), Substitution('optional_whitespace', generation_override=''), Substitution('item', label='item')]))]),
            'expression': Sequence([Substitution('sequence', label='choice'), Repetition(Sequence([Substitution('optional_whitespace'), Terminal('|'), Substitution('optional_whitespace'), Substitution('sequence', label='choice')]))]),

            # bnf script is sequence of production rules
            'main': Sequence([Substitution('rule', label='rule'), Repetition(Sequence([Substitution('optional_whitespace', generation_override='\n'), Substitution('rule', label='rule')]))]),
        })

        def parse_item(item: Node) -> Element:
            # there may or may not be a label
            labels = list(item.get('label', exclude=['contents']))
            label = labels[0].parsed_string if len(labels) == 1 else None
            # there must be exactly one contents node
            contents = next(item.get('contents'))
            node = contents.children[0]
            if node.rule == 'string':
                # TODO: actually parse the string, for escape chars and such
                return Terminal(next(node.get('contents')).parsed_string, label=label)
            elif node.rule == 'identifier':
                return Substitution(node.parsed_string, label=label)
            elif node.rule == 'option':
                return Option(parse_expression(next(node.get('expression'))), label=label)
            elif node.rule == 'repetition':
                return Repetition(parse_expression(next(node.get('expression'))), label=label)

        def parse_choice(sequence: Node) -> Sequence:
            children = list(map(parse_item, sequence.get('item')))
            return Sequence(children) if len(children) > 1 else children[0]

        def parse_expression(expression: Node) -> Element:
            children = list(map(parse_choice, expression.get('choice')))
            return Choice(children) if len(children) > 1 else children[0]

        def parse_rule(rule: Node) -> (name, Element):
            overrides = list(rule.get('override'))
            override = overrides[0].parsed_string if len(overrides) == 1 else None
            # there shoulddddd only be one, so just get the first
            name = next(rule.get('name')).parsed_string
            expression = next(rule.get('expression'))
            element = parse_expression(expression)
            element.generation_override = override
            return name, element

        def parse_rules(tree: Node) -> dict:
            return {name:element for name, element in list(map(parse_rule, tree.get('rule')))}

        # parse the source with the BNF grammar
        parsed = bnf.parse_complete(source.strip())
        rule_sets = map(parse_rules, parsed)
        grammars = map(cls, rule_sets)

        # just return the first valid grammar i supposed lol
        return next(grammars)

    @classmethod
    def load_bnf(cls, path: str):
        """Return a grammar described by an BNF file given the path to the file."""

        with open(path) as f:
            return cls.parse_bnf(f.read())

    def __str__(self):
        return '\n'.join(f'{rule_name} = {self.rules[rule_name]} .' for rule_name in self.rules)



from __future__ import annotations
from itertools import product, chain
from functools import reduce
import random
import operator
from .node import Node

def get_length(sequence) -> int:
    """Return the length of any sequence, which may be a generator.

    Beware of infinite sequences.
    """

    n = 0
    for _ in sequence:
        n += 1
    return n

def random_sample(sequence_func, n: int, length: int = None):
    """Specialized function to sample n items from a sequence, which may be a generator.

    sequence_func is a function to return the sequence.
    """

    # the predeturmined indices of the items to collect
    length = length or get_length(sequence_func())
    indices = random.sample(range(length), min(n, length))

    for i, item in enumerate(sequence_func()):
        if i in indices:
            yield item

def make_padding(indent: int) -> str:
    """Make a left-side whitespace padding for a given indention level."""

    return '   ' * indent

def label_node(sequence, label: str = None):
    """Optionally wrap all yielded nodes with a label node if provided."""

    for node in sequence:
        if label:
            yield Node(node.grammar, node.rule, node.string, node.region, [node], label=label)
        else:
            yield node

def parse_choice(elements: list, grammar: Grammar, rule: str, string, position: int = 0, verbose: bool = False):
    """Parse a choice among elements."""

    return chain(*(element.parse(grammar, rule, string, position, verbose) for element in elements))

def parse_sequence(elements: list, grammar: Grammar, rule: str, string, position: int = 0, verbose: bool = False):
    """Parse a sequence of elements."""

    def sequential_choices(elements: list, position: int) -> list:
        if elements:
            head, tail = elements[0], elements[1:]
            head_choices = head.parse(grammar, rule, string, position, verbose)
            for head_choice in head_choices:
                tail_choices = sequential_choices(tail, head_choice.region.stop)
                for tail_choice in tail_choices:
                    yield [head_choice] + tail_choice
        else:
            yield [Node(grammar, rule, string, slice(position, position))]

    for children in sequential_choices(elements, position):
        region = slice(children[0].region.start, children[-1].region.stop)
        yield Node(grammar, rule, string, region, children)

def generate_product(grammar: Grammar, generators_func, verbose: bool = False):
    """Generate the combinations of some generators.

    generators_func is a function to return a list of generators.
    """

    if verbose:
        length = get_length(generators_func())
        print("Generating product of {length} sequences.")
    lengths = list(map(get_length, generators_func()))
    length = reduce(operator.mul, lengths, 1)
    for prod in random_sample(lambda: product(*generators_func()), grammar.max_products, length):
        yield ''.join(prod)

    # TODO: get way more performance out of this lol

    #for i, prod in enumerate(product(*generators_func())):
    #    if i < grammar.max_products:
    #        if verbose:
    #            print(f'Joining sequence of length {len(prod)}.')
    #        yield ''.join(prod)
    #        if verbose:
    #            print(f'Finished joining sequence.')
    #    else:
    #        if verbose:
    #            print(f'Generation sequence too long; stopping generation here.')
    #        break



from __future__ import annotations
from dataclasses import dataclass, field

@dataclass
class Node:
    """Parse tree node."""

    # the grammar this was parsed with
    grammar: Grammar

    # the grammatical rule in the grammar that generated this node
    rule: str

    # the original string this was parsed from
    string: str

    # the region of the string parsed
    region: slice

    # a list of child nodes, if any
    children: list = field(default_factory=list)

    # label for this node for easy accessing in a list of children
    label: str = None

    def get(self, label, exclude=[]):
        """Yield all nearest branches with a given label.

        Will not traverse nodes with labels in the exclude list.
        """

        for child in self.children:
            if child.label == label:
                yield child
            elif child.label not in exclude:
                for match in child.get(label):
                    yield match

    @property
    def parsed_string(self) -> str:
        """Return the piece of string that was parsed by this node as a whole."""

        if self.children:
            return ''.join([child.parsed_string for child in self.children])
        else:
            return self.string[self.region]

    @property
    def is_empty(self) -> bool:
        """Return whether this node matches the empty string."""

        return self.parsed_string == ''

    @property
    def is_space(self) -> bool:
        """Return whether this node matches white space."""

        return self.parsed_string.isspace()

    @property
    def is_complete(self) -> bool:
        """Whether this node spans the whole string."""

        return self.region.start == 0 and self.region.stop == len(self.string)

    @property
    def filtered_children(self) -> list:
        """Return a list of children that are not empty or whitespace."""

        return list(filter(lambda child: not child.is_empty and not child.is_space, self.children))

    def __str__(self):
        label = f'({self.label!r}) ' if self.label else ''
        string = f'{self.rule} {label}= {self.parsed_string!r}'
        children = self.filtered_children
        for i, child_string in enumerate(map(str, children)):
            arrow, line = ('└ ', '  ') if i == len(children) - 1 else ('├ ', '│ ')
            body = f'\n{line}'.join(child_string.split('\n'))
            string += f'\n{arrow}{body}'
        return string



import setuptools

with open('README.md', 'r') as f:
    long_description = f.read()

setuptools.setup(
    name='nangram',
    version='0.0.4',
    author='Negative Nancy',
    author_email='negativefnnancy@gmail.com',
    description='Small little context-free expression generator and parser using user provided EBNF-like source.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/negativefnnancy/NanGram',
    packages=setuptools.find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.8',
)
